# SEC.gov Filing parsing

from typing import Tuple, Dict


class Filing:
    header = ""
    documents = None  # type: Dict[str, str]

    def __init__(self, filing: str):
        self.header, filing2 = self._extract_section(filing, "<SEC-HEADER>", "</SEC-HEADER>")

        if self.documents is None:
            self.documents = {}

        document, filing = self._extract_section(filing2, "<DOCUMENT>", "</DOCUMENT>")
        while len(document) > 0:
            tag = self._extract_tag(document, "<TYPE>")
            self.documents[tag] = document
            document, filing = self._extract_section(filing, "<DOCUMENT>", "</DOCUMENT>")

    def __repr__(self):
        ret = "Filings with\n"
        ret += " Header of Length: {0}\n".format(len(self.header))
        for key in self.documents:
            ret += " Document ({0}) of Length: {1}\n".format(key, len(self.documents[key]))
        return ret

    def _extract_section(self, filing: str, section_start: str, section_end: str) -> Tuple[str, str]:
        start = filing.find(section_start)
        end = filing.find(section_end)
        if start == -1 or end == -1:
            return "", filing  # not found
        section = filing[start + len(section_start) + 1 : end]
        filing = filing[end + len(section_end) + 1 :]
        return section, filing

    def _extract_tag(self, document: str, tag: str) -> str:
        start = document.find(tag)
        end = document.find("\n", start)
        if start == -1 or end == -1:
            return ""  # not found
        return document[start + len(tag) : end]



# SEC.GOV Edgar Interface

import httpx
import pandas as pd
from io import StringIO
import datetime
from typing import Tuple, Dict, Any, List
import json
import backoff


class edgar_index:

    # https://www.sec.gov/edgar/searchedgar/accessing-edgar-data.htm

    daily_index = "https://www.sec.gov/Archives/edgar/daily-index"
    full_index = "https://www.sec.gov/Archives/edgar/full-index"
    archives_url = "https://www.sec.gov/Archives"
    edgar_url = "https://www.sec.gov/Archives/edgar"

    _known_daily_index = {}  # type: Dict
    _known_full_index = {}  # type: Dict

    def _get_quarter(self, month: int):
        # Quarters are
        # QTR1 1-3
        # QTR2 4-6
        # QTR3 7-9
        # QTR4 10-12

        if month <= 3:
            return 1
        elif month <= 6:
            return 2
        elif month <= 9:
            return 3
        else:
            return 4

    def _get_full_index_url(self, date: datetime.datetime) -> str:
        quarter = self._get_quarter(date.month)
        return "{full_index}/{year}/QTR{quarter}/index.json".format(full_index=self.full_index, year=date.year, quarter=quarter)

    def _get_daily_index_url(self, date: datetime.datetime) -> str:
        quarter = self._get_quarter(date.month)
        return "{daily_index}/{year}/QTR{quarter}/index.json".format(daily_index=self.daily_index, year=date.year, quarter=quarter)

    def _get_full_index(self, date: datetime.datetime) -> Dict[str, Any]:
        quarter = self._get_quarter(date.month)
        key = "{year}_{quarter}".format(year=date.year, quarter=quarter)

        if key in self._known_full_index:
            return self._known_full_index[key]
        else:
            index = json.loads(httpx.get(self._get_full_index_url(date)).text)
            self._known_full_index[key] = index
            return index

    def _get_daily_index(self, date: datetime.datetime) -> Dict[str, Any]:
        quarter = self._get_quarter(date.month)
        key = "{year}_{quarter}".format(year=date.year, quarter=quarter)

        if key in self._known_daily_index:
            return self._known_daily_index[key]
        else:
            index = json.loads(httpx.get(self._get_daily_index_url(date)).text)
            self._known_daily_index[key] = index
            return index

    # ('https://www.sec.gov/Archives/edgar/full-index/2018/QTR1/master.idx', True)
    def _get_full_listing_url(self, date: datetime.datetime) -> Tuple[str, bool]:
        # master.idx
        file_name = "master.idx"

        index_dic = self._get_full_index(date)

        for item in index_dic["directory"]["item"]:
            if item["name"] == file_name:
                return ("{edgar_url}/{parent}{file_name}".format(edgar_url=self.edgar_url, parent=index_dic["directory"]["name"], file_name=file_name), True)

        return ("", False)

    # ('https://www.sec.gov/Archives/edgar/daily-index/2018/QTR1/master.20180102.idx',True)
    def _get_daily_listing_url(self, date: datetime.datetime) -> Tuple[str, bool]:
        # master.20180102.idx
        file_name = "master.{year}{month:02d}{day:02d}.idx".format(year=date.year, month=date.month, day=date.day)

        index_dic = self._get_daily_index(date)

        for item in index_dic["directory"]["item"]:
            if item["name"] == file_name:
                return ("{edgar_url}/{parent}{file_name}".format(edgar_url=self.edgar_url, parent=index_dic["directory"]["name"], file_name=file_name), True)

        return ("", False)

    # Gets the daily index of avalable filings
    # 98246|TIFFANY & CO|4|2018-01-12|edgar/data/98246/0000098246-18-000049.txt
    def get_full_listing(self, date: datetime.datetime) -> str:
        listing_url, status = self._get_full_listing_url(date)
        if status is False:
            return ""
        return httpx.get(listing_url).text

    def get_daily_listing(self, date: datetime.datetime) -> str:
        listing_url, status = self._get_daily_listing_url(date)
        if status is False:
            return ""
        return httpx.get(listing_url).text

    def get_full_listing_as_pd(self, date: datetime.datetime) -> pd.DataFrame:
        listing = self.get_full_listing(date)
        return self._format_listing_as_pd(listing, date)

    def get_daily_listing_as_pd(self, date: datetime.datetime) -> pd.DataFrame:
        listing = self.get_daily_listing(date)
        return self._format_listing_as_pd(listing, date)

    def _format_listing_as_pd(self, listing: str, date: datetime.datetime) -> pd.DataFrame:
        if len(listing) == 0:
            return None

        raw_rows = listing.splitlines()

        # validate header to be removed
        if not raw_rows[0].startswith("Description:"):
            raise Exception("Failed header check: Description")
        if not raw_rows[1].startswith("Last Data Received:"):
            raise Exception("Failed header check: Last Data Received")
        if not raw_rows[2].startswith("Comments:"):
            raise Exception("Failed header check: Comments")
        if not raw_rows[3].startswith("Anonymous FTP:"):
            raise Exception("Failed header check: Anonymous FTP")

        if raw_rows[4].startswith("Cloud HTTP:"):
            skip = 5
            while len(raw_rows[skip]) <= 1:
                skip += 1
            if not raw_rows[skip].startswith("CIK|Company Name|Form Type|Date Filed|Filename"):
                raise Exception("Failed header check: Cloud, Column Names")
            if not raw_rows[skip + 1].startswith("--------------------------------" + "------------------------------------------------"):
                raise Exception("Failed header check: Cloud, Data Seperator")
            skip += 2
        else:

            if not len(raw_rows[4]) <= 1:
                raise Exception("Failed header check: Line 5 not empty")
            if not raw_rows[5] == "CIK|Company Name|Form Type|Date Filed|File Name":
                raise Exception("Failed header check: Column names")
            if not raw_rows[6] == "--------------------------------------------------------------------------------":
                raise Exception("Failed header check: Data Seperator")
            skip = 7

        df = pd.read_csv(StringIO(listing), sep="|", index_col=False, skiprows=skip, names=["CIK", "Company Name", "Form Type", "Date Filed", "File Name"])
        df["Date Found"] = int("{year}{month:02d}{day:02d}".format(year=date.year, month=date.month, day=date.day))
        return df

    def get_filing_list(self, ciks: List[int], start_date: datetime.datetime, end_date: datetime.datetime) -> pd.DataFrame:
        df = None
        for day in range((end_date - start_date).days + 1):
            date = start_date + datetime.timedelta(days=day)
            if date.isoweekday() <= 5:
                df_test = self.get_daily_listing_as_pd(date)
                if df_test is None:
                    continue

                df_test = df_test.loc[df_test["CIK"].isin(ciks)]
                if df is None:
                    df = df_test
                else:
                    df = pd.concat([df, df_test], ignore_index=True)
        return df

    @backoff.on_exception(backoff.expo, (httpx.exceptions.NetworkError))
    def get_filing(self, url: str):
        return httpx.get("{archives_url}/{url}".format(archives_url=self.archives_url, url=url)).text



import datetime
import pandas as pd
import numpy as np
from edgar3 import edgar_index


def get_13f_listings(date: datetime.datetime, populate: bool) -> pd.DataFrame:
    ed_i = edgar_index.edgar_index()
    df = ed_i.get_full_listing_as_pd(date)
    df = df[df["Form Type"].isin(["13F-HR", "13F-HR/A", "13F-NT", "13F-NT/A"])].reset_index(drop=True)
    df["File"] = np.nan
    if populate:
        for index, row in df.iterrows():
            df["File"] = ed_i.get_filing(row["File Name"])
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edgar3']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.10.0,<2.0.0', 'httpx>=0.11.0,<0.12.0', 'pandas>=0.25.3,<0.26.0']

setup_kwargs = {
    'name': 'edgar3',
    'version': '1.1.8',
    'description': 'Simple 13F Edgar Extractor',
    'long_description': None,
    'author': 'Ken Farr',
    'author_email': 'ken@farr.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .filing import Filing
import xml.etree.ElementTree as ET
from typing import Dict


class Filing_13F(Filing):
    def __init__(self, filing: str):
        super().__init__(filing)

    def __repr__(self):
        ret = f"13F Filings for {self.manager_name}({self.cik})\n" f" Period: {self.period_of_report}\n" f" Reported: {self.signature_date}\n" f" Header of Length: {0}\n".format(len(self.header))
        for key in self.documents:
            ret += " Document ({0}) of Length: {1}\n".format(key, len(self.documents[key]))
        return ret

    def process_information_table(self):
        try:
            document = self.documents["INFORMATION TABLE"]
        except KeyError:
            return False

        xml_doc, ext = self._extract_section(document, "<XML>", "</XML>")
        if len(xml_doc) == 0:
            return False
        root = ET.fromstring(xml_doc)
        namespace = {"ns1": "http://www.sec.gov/edgar/document/thirteenf/informationtable"}
        self.holdings = []
        for child in root:
            try:
                holding = Holding(child, namespace)
                self.holdings.append(holding)
            except ValueError as error:
                print("Error processing Holding:", error)

        return True

    def _process_header_data(self, root: ET.Element, namespace: Dict[str, str]):
        header_data = _get_element(root, "ns1:headerData", namespace)
        filer_info = _get_element(header_data, "ns1:filerInfo", namespace)
        self.period_of_report = _get_element_text(filer_info, "ns1:periodOfReport", namespace)

        filer = _get_element(filer_info, "ns1:filer", namespace)
        creds = _get_element(filer, "ns1:credentials", namespace)
        self.cik = _get_element_text(creds, "ns1:cik", namespace)

    def _process_form_data(self, root: ET.Element, namespace: Dict[str, str]):
        form_data = _get_element(root, "ns1:formData", namespace)
        cover_page = _get_element(form_data, "ns1:coverPage", namespace)
        filing_manager = _get_element(cover_page, "ns1:filingManager", namespace)
        self.manager_name = _get_element_text(filing_manager, "ns1:name", namespace)

        address = _get_element(filing_manager, "ns1:address", namespace)
        self.street1 = _get_element_text(address, "ns2:street1", namespace)
        self.street2 = _get_element_text(address, "ns2:street2", namespace, "")
        self.city = _get_element_text(address, "ns2:city", namespace)
        self.state_or_country = _get_element_text(address, "ns2:stateOrCountry", namespace)
        self.zip_code = _get_element_text(address, "ns2:zipCode", namespace)

        signature_block = _get_element(form_data, "ns1:signatureBlock", namespace)
        self.signature_date = _get_element_text(signature_block, "ns1:signatureDate", namespace)

    def process_13f_hr(self):
        try:
            document = self.documents["13F-HR"]
        except KeyError:
            return False

        xml_doc, ext = self._extract_section(document, "<XML>", "</XML>")
        if len(xml_doc) == 0:
            return False
        root = ET.fromstring(xml_doc)
        namespace = {"ns1": "http://www.sec.gov/edgar/thirteenffiler", "ns2": "http://www.sec.gov/edgar/common"}
        self._process_header_data(root, namespace)
        self._process_form_data(root, namespace)
        return True


def _get_element(root: ET.Element, path: str, namespace: Dict[str, str]) -> ET.Element:
    x = root.find(path, namespace)
    if x is None:
        raise ValueError(f"During processing of a 13f, the {path} element was not found on the expected parent")
    return x


def _get_element_text(root: ET.Element, path: str, namespace: Dict[str, str], default: str = None):
    x = root.find(path, namespace)
    if x is None:
        if default is None:
            raise ValueError(f"During processing a 13f, the {path} element was not found on the expected parent, thus no .text")
        else:
            return default
    return x.text


class Holding:
    def __init__(self, root: ET.Element, namespace: Dict[str, str]):
        try:
            self.nameOfIssuer = _get_element_text(root, "ns1:nameOfIssuer", namespace, "")
            self.titleOfClass = _get_element_text(root, "ns1:titleOfClass", namespace, "")
            self.cusip = _get_element_text(root, "ns1:cusip", namespace, "")
            self.value = int(float(_get_element_text(root, "ns1:value", namespace, "0"))) * 1000

            shares = root.find("ns1:shrsOrPrnAmt", namespace)

            # not sure if there's a way for additional shares types to be encoded here
            # so we will throw an error if it's not 2
            if shares is None or len(shares) != 2:
                raise ValueError(self.nameOfIssuer + ": shrsOrPrnAmt != 2")

            self.number = int(float(_get_element_text(shares, "ns1:sshPrnamt", namespace, "0")))
            self.type = _get_element_text(shares, "ns1:sshPrnamtType", namespace, "")

        except ValueError:
            raise ValueError(self.nameOfIssuer)
        except AttributeError:
            raise ValueError(self.nameOfIssuer)

    def __repr__(self):
        ret = self.nameOfIssuer + "(${0}:{1} @ {2})".format(self.value, self.number, self.value / self.number)
        return ret


#
# <ns1:informationTable xmlns:ns1="http://www.sec.gov/edgar/document/thirteenf/informationtable">
# 	<ns1:infoTable>
#         <ns1:nameOfIssuer>AMERICAN EXPRESS CO </ns1:nameOfIssuer>
# 		<ns1:titleOfClass>COM</ns1:titleOfClass>
# 		<ns1:cusip>025816109</ns1:cusip>
# 		<ns1:value>27055</ns1:value>
# 		<ns1:shrsOrPrnAmt>
# 			<ns1:sshPrnamt>272430</ns1:sshPrnamt>
# 			<ns1:sshPrnamtType>SH</ns1:sshPrnamtType>
# 		</ns1:shrsOrPrnAmt>
# 		<ns1:investmentDiscretion>SOLE</ns1:investmentDiscretion>
# 		<ns1:votingAuthority>
# 			<ns1:Sole>272430</ns1:Sole>
# 			<ns1:Shared>0</ns1:Shared>
# 			<ns1:None>0</ns1:None>
# 		</ns1:votingAuthority>
# 	</ns1:infoTable>"""
#
